
'''

Solve a Poisson transmission problem 

- \Delta p_top    = f_top      in \Omega_top    := (0,1)x(0.5, 1)
- \Delta p_bottom = f_bottom   in \Omega_bottom := (0,1)x(0, 0.5)

                        p_top - p_bottom = jump1                on interface
         \nabla p_top \cdot n - \nabla p_bottom \cdot n = jump2 on interface 

with a hybrid mixed FEM using the "skeleton" programming style of NGSolve.

The exact solution is: 

    p_top    = sin(pi*x)*sin(pi*y) 
and 
    p_bottom = 8*x*(1-x)*y*(1-y)

'''

# The "skeleton" version works fine

from ngsolve import *
from netgen.geom2d import SplineGeometry

from numpy import pi

ngsglobals.msg_level = 1


# Different sets of: Exact solution + source terms 

# Case 1

p_top    = sin(pi*x)*sin(pi*y)
p_bottom = 8*x*(1-x)*y*(1-y)
ftop     = 2*pi**2*p_top
fbottom  = 16*y*(1-y) + 16*x*(1-x)
gradp_top    = CoefficientFunction( (pi*cos(pi*x)*sin(pi*y), pi*sin(pi*x)*cos(pi*y)) )
gradp_bottom = CoefficientFunction( ( 8*(1-2*x)*y*(1-y), 8*x*(1-x)*(1-2*y) ) )

# Case 2

""" p_top    = sin(pi*x)*sin(pi*y)
p_bottom = cos(pi*x)*cos(pi*y)
ftop     = 2*pi**2*p_top
fbottom  = 2*pi**2*p_bottom
gradp_top    = CoefficientFunction( (pi*cos(pi*x)*sin(pi*y), pi*sin(pi*x)*cos(pi*y)) )
gradp_bottom    = -CoefficientFunction( (pi*sin(pi*x)*cos(pi*y), pi*cos(pi*x)*sin(pi*y)) ) """

# Case 3

""" p_top = cos(pi*x)*cos(pi*y)
p_bottom = -p_top
ftop = 2*pi**2*p_top
fbottom = -ftop
gradp_top = -CoefficientFunction( (pi*sin(pi*x)*cos(pi*y), pi*cos(pi*x)*sin(pi*y)) )
gradp_bottom = -gradp_top """

# Case 4

""" p_top = atan((y+0.5)/(x+0.5))
p_bottom = -p_top
ftop = 0
fbottom = -ftop
gradp_top    = CoefficientFunction( (- (y+0.5)/( (x+0.5)**2 + (y+0.5)**2 ) , (x+0.5)/((x+0.5)**2 + (y+0.5)**2)) )
gradp_bottom = -gradp_top """

# jumps on interface

jump1 = p_top - p_bottom
jump2 = gradp_top - gradp_bottom

# Geometry and mesh

def topBottom():
    geometry = SplineGeometry()
    pnts     = [ (0, 0), (1, 0),  (1, 0.5), (1, 1), (0, 1), (0, 0.5)]
    pnums    = [geometry.AppendPoint(*p) for p in pnts]
    # start-point, end-point, boundary-condition, domain on left side, domain on right side:
    lines = [ (pnums[0], pnums[1], "gammaBottom",  2, 0),
              (pnums[1], pnums[2], "gammaBottom",  2, 0), 
              (pnums[2], pnums[3], "gammaTop",     1, 0),
              (pnums[3], pnums[4], "gammaTop",     1, 0),
              (pnums[4], pnums[5], "gammaTop",     1, 0),
              (pnums[5], pnums[0], "gammaBottom",  2, 0),
              (pnums[5], pnums[2], "interface",    1, 2)]
    for p1,p2,bc,left,right in lines:
        geometry.Append( ["line", p1, p2], bc=bc, leftdomain=left, rightdomain=right)

    geometry.SetMaterial(1, "top")
    geometry.SetMaterial(2, "bottom")
        
    return geometry

mesh = Mesh(topBottom().GenerateMesh (maxh=0.2))

pExact = CoefficientFunction([p_top if mat == "top" else p_bottom for mat in mesh.GetMaterials()])
source = CoefficientFunction([ftop if mat == "top" else fbottom for mat in mesh.GetMaterials()])

n = specialcf.normal(mesh.dim)

# Mixed finite element spaces

order = 4

V = HDiv(mesh, order = order, discontinuous=True)
M = L2(mesh,   order = order-1)
L  = FacetFESpace(mesh, order=order,  dirichlet="gammaTop|gammaBottom")

X = FESpace([V, M,  L], dgjumps = True)
u,  p,  lmbda = X.TrialFunction()
v,  q,  mu    = X.TestFunction()


jump_u = (u-u.Other())*n
jump_v = (v-v.Other())*n

mean_un = 0.5*(u+u.Other())
mean_v = 0.5*(v+v.Other())


# Bilinear and linear forms

a = BilinearForm(X, symmetric=True)#, condense=True)
a += u*v*dx + div(u)*q*dx + div(v)*p*dx 
a += (-lmbda*jump_v - mu*jump_u) * dx(skeleton=True)

#a += -1e-12*(lmbda)*(mu)*dx(element_boundary=True)

# Right-hand side

f = LinearForm(X)
f += -source*q*dx # Source
f +=pExact*(v*n)*ds(skeleton=True, definedon=mesh.Boundaries("gammaTop|gammaBottom")) # Dirichlet BC (natural)
f += jump1*(mean_v*n)*ds(skeleton=True, definedon=mesh.Boundaries("interface")) # First transmission condition
f +=  -(jump2*n)*mu*ds(skeleton=True, definedon=mesh.Boundaries("interface"))   # Second transmission condition


#Solving the problem
a.Assemble()
f.Assemble()

gfu = GridFunction(X)
gfu.components[2].Set(pExact, BND, definedon=mesh.Boundaries("gammaTop|gammaBottom")) # Dirichlet BC for lmbda (essential)

r = f.vec.CreateVector()
r.data = f.vec - a.mat*gfu.vec
gfu.vec.data += a.mat.Inverse(freedofs=X.FreeDofs(), inverse="umfpack")*r 



gfp = gfu.components[1]  

# L2 error

upe = GridFunction(M, name="exact") 
upe.Set(pExact) # Projection of the exact solution on the FE space M

err = sqrt(Integrate((upe-gfp)*(upe-gfp),mesh)) 
print(err)

# Drawing 
Draw(upe)
Draw(gfp)